// CsEntity.h

#pragma once

#ifndef _CS_ENTITY_H_
#define _CS_ENTITY_H_



#include "hash.h"
//#include "CsScopeTable.h"
#include "SingleList.h"
#include "CsEntityVisitor.h"


using namespace HashString;

struct CsNode;
struct CsConstructorDeclaration;
struct CsDestructorDeclaration;
struct CsConstantDeclarator;
struct CsVariableDeclarator;
struct CsMethodDeclaration;
struct CsOperatorDeclaration;
struct CsConversionOperatorDeclaration;
struct CsPropertyDeclaration;
struct CsDelegate;
struct CsEventDeclaration;
struct CsEventDeclarator;
struct CsEventAccessor;
struct CsEnum;
struct CsBlock;
struct CsLocalConstantDeclarator;
struct CsLocalVariableDeclarator;
struct CsStatement;
struct CsFormalParameter;
struct CsFormalParameterList;
struct CsUsingDirective;
struct CsLabel;

struct CsEntityInterface;
struct CsEntityProperty;
struct CsEntityEvent;
struct CsEntityBlock;
struct CsEntityArraySpecifier;
struct CsEntityMethodSpecifier;

struct ClsAssembly;

typedef ScopeHashTable CsScopeTable;

//#undef new
//#undef delete

//------------------------------------------------------------------------------


//------------------------------------------------------------------------------

#define DEFINE_ENTITY_ALLOCATOR(Type)\
inline void * Type::operator new ( size_t size )\
{\
	return FixedSmallObjAllocator < ALIGN_TO_2_POW_N_OR_64 ( sizeof ( Type ) ), EntityAllocatorId, OwnPageAllocator >::allocate();\
}\
\
inline void Type::operator delete ( void * address )\
{\
	FixedSmallObjAllocator < ALIGN_TO_2_POW_N_OR_64 ( sizeof ( Type ) ), EntityAllocatorId, OwnPageAllocator >::deallocate ( address );\
}\

//------------------------------------------------------------------------------
// cs_entity
//------------------------------------------------------------------------------

enum cs_entity
{
// global entities
	ent_class,
	ent_struct,
	ent_interface,
	ent_enum,
	ent_delegate,
	ent_namespace,
	ent_constant,
	ent_variable,
	ent_method,
	ent_property,
	ent_event,
// mix entities
	ent_formal_parameter,
	ent_formal_parameter_list,
// local entities
	ent_block_variable,
	ent_block,
	ent_local_constant,
	ent_local_variable,
	ent_end,
};

//------------------------------------------------------------------------------
// cs_entity_access
//------------------------------------------------------------------------------

enum cs_entity_access
{
	ac_compiler_controlled,
	ac_public,
	ac_protected,
	ac_private,
	ac_internal,
	ac_protected_internal,
	ac_protected_and_internal,
};

//------------------------------------------------------------------------------
// entity_flags
//------------------------------------------------------------------------------

enum entity_flags
{
	f_error                       = 0x0001, // this flag used to skip entity in lookup algorithms
	f_generated                   = 0x0002, // entity generated by parser (for example: methods Invoke(), BeginInvoke(), EndInvoke() for delegates)
	f_compiler_attribute          = 0x0004, // compiler attribute ( like System.Runtime.CompilerServices.IndexerName )
	f_varargs                     = 0x0008, // method accepts an arbitrary number of arguments ( __arglist last formal parameter )
// temple flags (for local variables and constants only)
	ft_declared                   = 0x4000,
	ft_used_before_declared       = 0x8000,

	f_metadatamask                = 0xfffffff8,

// 22.1.4 Flags for Events [EventAttributes]

	f_event_SpecialName           = 0x0200, // Event is special
	f_event_RTSpecialName         = 0x0400, // CLI provides 'special' behavior, depending upon the name of the event

// 22.1.5 Flags for Fields [FieldAttributes]

	f_field_Static                = 0x0010, // Defined on type, else per instance
	f_field_InitOnly              = 0x0020, // Field may only be initialized, not written to after init
	f_field_Literal               = 0x0040, // Value is compile time constant
	f_field_NotSerialized         = 0x0080, // Field does not have to be serialized when type is remoted
// Additional flags
	f_field_HasFieldRVA           = 0x0100, // Field has RVA 
//
	f_field_SpecialName           = 0x0200, // Field is special
// Additional flags
	f_field_RTSpecialName         = 0x0400, // CLI provides 'special' behavior, depending upon the name of the field
	f_field_HasFieldMarshal       = 0x1000, // Field has marshalling information
// Interop Attributes
	f_field_PInvokeImpl           = 0x2000, // Implementation is forwarded through PInvoke
// Additional flags
	f_field_HasDefault            = 0x8000, // Field has default 

//	22.1.9 Flags for Methods [MethodAttributes]

// Interop attributes
	f_method_UnmanagedExport      = 0x0008, // Reserved: shall be zero for conforming implementations 
// 
	f_method_Static               = 0x0010, // Defined on type, else per instance
	f_method_Final                = 0x0020, // Method may not be overridden
	f_method_Virtual              = 0x0040, // Method is virtual
	f_method_HideBySig            = 0x0080, // Method hides by name+sig, else just by name
//
	f_method_VtableLayoutMask     = 0x0000, // Use this mask to retrieve vtable attributes
	f_method_ReuseSlot            = 0x0000, // Method reuses existing slot in vtable
	f_method_NewSlot              = 0x0100, // Method always gets a new slot in the vtable
//
	f_method_Abstract             = 0x0400, // Method does not provide an implementation
	f_method_SpecialName          = 0x0800, // Method is special
// Additional flags
	f_method_RTSpecialName        = 0x1000, // CLI provides 'special' behavior, depending upon the name of the method
// Interop attributes
	f_method_PInvokeImplementation = 0x2000, // Implementation is forwarded through PInvoke 
// Interop attributes
	f_method_HasSecurity          = 0x4000, // Method has security associate with it
	f_method_RequireSecObject     = 0x8000, // Method calls another method containing security code

// 22.1.14 Flags for Types [TypeAttributes]

// Class layout attributes
	f_type_LayoutMask             = 0x000018, // Use this mask to retrieve class layout information
	f_type_AutoLayout             = 0x000000, // Class fields are auto-laid out
	f_type_SequentialLayout       = 0x000008, // Class fields are laid out sequentially
	f_type_ExplicitLayout         = 0x000010, // Layout is supplied explicitly
// Class semantics attributes
	f_type_ClassSemanticsMask     = 0x000020, // Use this mask to retrive class semantics information
	f_type_Class                  = 0x000000, // Type is a class
	f_type_Interface              = 0x000020, // Type is an interface
// Special semantics in addition to class semantics
	f_type_Abstract               = 0x000080, // Class is abstract
	f_type_Sealed                 = 0x000100, // Class cannot be extended
	f_type_SpecialName            = 0x000400, // Class name is special
// Additional Flags
	f_type_RTSpecialName          = 0x000800, // CLI provides 'special' behavior, depending upon the name of the Type
// Implementation Attributes
	f_type_Import                 = 0x001000, // Class/Interface is imported
	f_type_Serializable           = 0x002000, // Class is serializable
// String formatting Attributes
	f_type_StringFormatMask       = 0x030000, // Use this mask to retrieve string information for native interop
	f_type_AnsiClass              = 0x000000, // LPSTR is interpreted as ANSI
	f_type_UnicodeClass           = 0x010000, // LPSTR is interpreted as Unicode
	f_type_AutoClass              = 0x020000, // LPSTR is interpreted automatically
// Additional Flags
	f_type_HasSecurity            = 0x040000, // Type has security associate with it
// Class Initialization Attributes
	f_type_BeforeFieldInit        = 0x100000, // Initialize the class before first static field access
};

//------------------------------------------------------------------------------
// cs_entity_type
//------------------------------------------------------------------------------

enum cs_entity_type
{
// types defined in metadata
	et_unknown,
	et_void,
	et_boolean,
	et_char,
	et_int8,
	et_uint8,
	et_int16,
	et_uint16,
	et_int32,
	et_uint32,
	et_int64,
	et_uint64,
	et_float32,
	et_float64,
	et_string,
	et_ptr, // pointer to type
	et_byref, // reference to type
	et_valuetype, // struct
	et_class,
	et_array, // elements array of type
	et_typedbyref, // System.Type
	et_intptr,
	et_uintptr,
	et_fnptr, // pointer to method
	et_object,
// additional types introduced to simplify C# parser logic
// these types maybe defined only from C# source files
	et_literal_null, // used for implicit conversion from null to other types
	//et_literal_0, // used for implicit conversion from literal 0 to any enum
	//et_constant_expression, // prefix, used for implicit conversion from constant expression to other integral types
	et_arglist, // special stuff type for __arglist

	et_error, // used for catch erroneous operators
};

//------------------------------------------------------------------------------
// cs_entity_method_calling_convention
//------------------------------------------------------------------------------

//enum cs_entity_method_calling_convention
//{
//	ecc_none,
//	ecc_none,
//	ecc_none,
//	ecc_none,
//	ecc_none,
//	ecc_none,
//};

//------------------------------------------------------------------------------
// cs_entity_type_specifier
//------------------------------------------------------------------------------

enum cs_entity_type_specifier
{
	ets_empty,
	ets_pointer,
	ets_array, // +szarray?
};

//------------------------------------------------------------------------------
// cs_formal_parameter_kind
//------------------------------------------------------------------------------

enum cs_formal_parameter_kind
{
	fpk_empty = 0,
	fpk_in,
	fpk_out,
	fpk_inout,
};

//------------------------------------------------------------------------------
// DeletedEntityInfo
//------------------------------------------------------------------------------


//------------------------------------------------------------------------------
// CsEntity
//------------------------------------------------------------------------------

struct CsEntity
{
	CsEntity ( cs_entity type_, cs_entity_access ac, PHashString name_ )
		:	name ( name_ )
		, e ( type_ )
		, access ( ac )
		, flags1 ( 0 )
		, flags2 ( 0 )
		, parent(0)
		, assembly(0)
	{
	}

	CsEntity ( cs_entity type_, cs_entity_access ac )
		: e ( type_ )
		, access ( ac )
		, flags1 ( 0 )
		, flags2 ( 0 )
		, parent(0)
		, assembly(0)
	{
	}

	virtual ~CsEntity()
	{
		//XTRACE1 ( "entity delete %p\n", this );
		//XASSERT ( id != 2427556 );
	}

	virtual void visit ( CsEntityVisitor * visitor ) =0;
	//{
	//	XASSERT(0);
	//}


	virtual bool isSharedEntity() const
	{
		return assembly != 0;
	}

	inline bool isNamespaceOrType() const
	{
		return e == ent_class
			|| e == ent_struct
			|| e == ent_interface
			|| e == ent_enum
			|| e == ent_delegate
			|| e == ent_namespace;
	}

	inline bool isType() const
	{
		return e == ent_class
			|| e == ent_struct
			|| e == ent_interface
			|| e == ent_enum
			|| e == ent_delegate;
	}

	inline bool isNamespace() const
	{
		return e == ent_namespace;
	}

	inline bool isIdentifier() const
	{
		return 
			   e == ent_constant
			|| e == ent_variable
			|| e == ent_property
			|| e == ent_event
			|| e == ent_local_constant
			|| e == ent_local_variable
			|| e == ent_formal_parameter
			|| e == ent_class
			|| e == ent_struct
			|| e == ent_interface
			|| e == ent_enum
			|| e == ent_delegate
			|| e == ent_namespace
			|| e == ent_method;
	}

	inline bool isField() const
	{
		return 
			   e == ent_constant
			|| e == ent_variable
			|| e == ent_property
			|| e == ent_event;
	}

	inline bool isMethod() const
	{
		return e == ent_method;
	}

	inline bool isVirtualMethod() const
	{
		return ( flags2 & f_method_Virtual ) != 0;
	}

	inline bool isOverrideMethod() const
	{
		return ( flags2 & f_method_Virtual ) && !(flags2 & f_method_NewSlot);
	}

	inline static bool isParentChild ( CsEntity * parent, CsEntity * child )
	{
		CsEntity * child_parent = child->parent;
		while ( child_parent )
		{
			if ( parent == child_parent )
				return true;
			child_parent = child_parent->parent;
		}
	}




public:
	cs_entity e : 8;
	cs_entity_access access : 8;
	unsigned __int16 flags1;
	PHashString name;
	unsigned __int32 flags2;
	CsEntity * parent;
	ClsAssembly * assembly;
};

//------------------------------------------------------------------------------
// CsScopeEntity
//------------------------------------------------------------------------------

struct CsScopeEntity : public CsEntity
{
public:
	CsScopeEntity ( cs_entity ent, cs_entity_access ac, WPHashString name, CsNode * node_ )
		: CsEntity ( ent, ac, name )
		, node(node_)
	{}

	CsScopeEntity ( cs_entity ent, cs_entity_access ac )
		: CsEntity ( ent, ac )
		, node(0)
	{}

	CsScopeEntity ( cs_entity ent, cs_entity_access ac, WPHashString name, CsNode * node_, unsigned power_of_2 )
		: CsEntity ( ent, ac, name )
		, scope_table ( power_of_2 )
		, node(node_)
	{}

	virtual ~CsScopeEntity()
	{
		scope_table.delete_entities();
	}

	virtual void visit ( CsEntityVisitor * visitor );


public:
	CsScopeTable scope_table;
	CsNode * node;
};

//------------------------------------------------------------------------------
// CsEntityNamespace
//------------------------------------------------------------------------------

struct CsEntityNamespace : public CsScopeEntity
{
	CsEntityNamespace ( WPHashString phs )
		: CsScopeEntity ( ent_namespace, ac_public, phs, 0 )
	{}

	CsEntityNamespace ( cs_entity type, cs_entity_access ac )
		: CsScopeEntity ( type, ac )
	{}

	virtual ~CsEntityNamespace()
	{}

	virtual void visit ( CsEntityVisitor * visitor )
	{
		visitor->visit ( this );
		CsScopeEntity::visit ( visitor );
	}

	bool deleteProjectEntities();

	static inline void * operator new ( size_t size );
	static inline void operator delete ( void * address );
};

DEFINE_ENTITY_ALLOCATOR(CsEntityNamespace)

//------------------------------------------------------------------------------
// DeletedCsEntityTypeRefInfo
//------------------------------------------------------------------------------


//------------------------------------------------------------------------------
// DeletedCsEntityArraySpecifierInfo
//------------------------------------------------------------------------------


//------------------------------------------------------------------------------
// CsEntityTypeRef
//------------------------------------------------------------------------------

struct CsEntityTypeRef
{
	CsEntityTypeRef()
		: type ( et_unknown )
		, bConstantExpression ( false )
		, bZero ( false )
//#if CS_PARSER_TEST
//		, id ( st_id++ )
//#endif // CS_PARSER_TEST
	{
		u.entity = 0;
//#if CS_PARSER_TEST
//		if ( st_bCollectCreatedObjectsInfo )
//			st_CreatedObjectsInfo.push_back
//				( DeletedCsEntityTypeRefInfo ( id ) );
//#endif
	}

	//explicit CsEntityTypeRef ( const CsEntityTypeRef & rhs )
	//	: type ( rhs.type )
	//  , entity ( rhs.entity )
	//{}

	~CsEntityTypeRef()
	{
	}

	static inline void * operator new ( size_t size );
	static inline void operator delete ( void * address );




// clear CsEntityTypeRef to avoid double delete
	inline void clear_without_delete()
	{
		type = et_unknown;
	}

	void delete_entities();
	CsEntityTypeRef clone() const;

	inline void clear()
	{
		type = et_unknown;
		bConstantExpression = false;
		bZero = false;
		u.entity = 0;
	}

	bool equal ( const CsEntityTypeRef & rhs ) const;


	cs_entity_type type : 8;
	bool bConstantExpression;
	bool bZero;
	union
	{
		CsEntity * entity;
		CsEntityTypeRef * type;
		CsEntityArraySpecifier * array_specifier;
		CsEntityMethodSpecifier * method_specifier;
	}
	u;

	bool isValueType() const
	{
		switch ( type )
		{
			case et_boolean:
			case et_char:
			case et_int8:
			case et_uint8:
			case et_int16:
			case et_uint16:
			case et_int32:
			case et_uint32:
			case et_int64:
			case et_uint64:
			case et_float32:
			case et_float64:
			case et_valuetype:
				return true;
			default:
				return false;
		}
	}

	bool isVoidOrValueType() const
	{
		return isValueType()
			|| ( type == et_void );
	}

	bool isReferenceType() const
	{
		switch ( type )
		{
			case et_string:
			case et_byref:
			case et_class:
			case et_array:
			case et_typedbyref:
			case et_object:
				return true;
		}
		return false;
	}

};

DEFINE_ENTITY_ALLOCATOR(CsEntityTypeRef)

//------------------------------------------------------------------------------
// CsEntityArrayBoundSpecifier
//------------------------------------------------------------------------------

struct CsEntityArrayBoundSpecifier
{
	CsEntityArrayBoundSpecifier()
		: size(0)
		, low_bound(0)
	{}

	bool equal ( const CsEntityArrayBoundSpecifier & rhs ) const
	{
		return size == rhs.size && low_bound == rhs.low_bound;
	}

	int size;
	int low_bound;
};

//------------------------------------------------------------------------------
// CsEntityArraySpecifier
//------------------------------------------------------------------------------

struct CsEntityArraySpecifier
{
	CsEntityArraySpecifier()
		: rank(0)
		, bounds(0)
	{
		//TRACE1("CsEntityArraySpecifier::create %p\n", this );
	}

	CsEntityArraySpecifier ( int rank_ ) // special for ELEMENT_TYPE_SZARRAY
		: rank(rank_)
	{
		bounds = new CsEntityArrayBoundSpecifier[rank_];
		//TRACE1("CsEntityArraySpecifier::create %p\n", this );
	}

	~CsEntityArraySpecifier()
	{
		type.delete_entities();
		delete[] bounds;
		//TRACE1("CsEntityArraySpecifier::delete %p\n", this );
	}

	void stop_by_id()
	{
		//XASSERT ( id != 106054 );
	}

	CsEntityArraySpecifier * clone() const;

	bool equal ( const CsEntityArraySpecifier & rhs ) const;
	bool equal_bounds ( const CsEntityArraySpecifier & rhs ) const;



	CsEntityTypeRef type;
	int rank;
	CsEntityArrayBoundSpecifier * bounds;
};

//------------------------------------------------------------------------------
// CsEntityFormalParameterSpecifier
//------------------------------------------------------------------------------

struct CsEntityFormalParameterSpecifier
{
	CsEntityFormalParameterSpecifier()
		: kind ( fpk_empty )
		, bSystemParamArrayAttribute ( false )
	{}

	~CsEntityFormalParameterSpecifier()
	{}

	void delete_entities()
	{
		type.delete_entities();
	}

	CsEntityFormalParameterSpecifier clone() const
	{
		CsEntityFormalParameterSpecifier ret_value;
		ret_value.type = this->type.clone();
		ret_value.kind = this->kind;
		return ret_value;
	}

	bool equal ( const CsEntityFormalParameterSpecifier & rhs ) const
	{
		return type.equal ( rhs.type );
	}


	CsEntityTypeRef type;
	cs_formal_parameter_kind kind : 8;
	bool bSystemParamArrayAttribute;
};

//------------------------------------------------------------------------------
// CsEntityFormalParametersSpecifier
//------------------------------------------------------------------------------

struct CsEntityFormalParametersSpecifier
{
	CsEntityFormalParametersSpecifier()
		: n_params(0)
		, params(0)
	{}

	void delete_entities()
	{
		if ( params )
		{
			for ( int i = 0; i != n_params; ++i )
				params[i].delete_entities();
			delete[] params;
		}
	}

	~CsEntityFormalParametersSpecifier()
	{}

	CsEntityFormalParametersSpecifier clone() const;

	bool equal ( const CsEntityFormalParametersSpecifier & rhs ) const;


	int n_params;
	CsEntityFormalParameterSpecifier * params;

	typedef CsEntityFormalParametersSpecifier inherited;
};

//------------------------------------------------------------------------------
// CsEntityBlock
//------------------------------------------------------------------------------

struct CsEntityBlock : public CsScopeEntity
{
	CsEntityBlock ( CsNode * decl_ )
		: CsScopeEntity ( ent_block, ac_public )
		, decl(decl_)
		, blocks(0)
	{}

	virtual ~CsEntityBlock()
	{
		if ( blocks )
		{
			blocks->delete_nodes_and_values();
			delete blocks;
		}
	}

	virtual void visit ( CsEntityVisitor * visitor )
	{
		visitor->visit ( this );
		CsScopeEntity::visit ( visitor );
	}

	static inline void * operator new ( size_t size );
	static inline void operator delete ( void * address );

	CsNode * decl;
	single_list1<CsEntity*> * blocks; // list of pointers to nested blocks, list used only nested entities
};

DEFINE_ENTITY_ALLOCATOR(CsEntityBlock)

//------------------------------------------------------------------------------
// CsEntityFormalParameter
//------------------------------------------------------------------------------

struct CsEntityFormalParameter : public CsEntity
{
	CsEntityFormalParameter ( WPHashString phs, CsFormalParameter * param_ )
		: CsEntity ( ent_formal_parameter, ac_public, phs )
		, param ( param_ )
		, kind ( fpk_empty )
		, bSystemParamArrayAttribute ( false )
	{}

	virtual ~CsEntityFormalParameter()
	{
		type.delete_entities();
	}

	virtual void visit ( CsEntityVisitor * visitor )
	{
		visitor->visit ( this );
	}

	static inline void * operator new ( size_t size );
	static inline void operator delete ( void * address );

	CsEntityTypeRef type;
	cs_formal_parameter_kind kind : 8;
	bool bSystemParamArrayAttribute;
	CsFormalParameter * param;
};

DEFINE_ENTITY_ALLOCATOR(CsEntityFormalParameter)

//------------------------------------------------------------------------------
// CsEntityFormalParameterList
//------------------------------------------------------------------------------

struct CsEntityFormalParameterList : public CsEntity
{
	CsEntityFormalParameterList ( CsFormalParameterList * param_list_ )
		: CsEntity ( ent_formal_parameter_list, ac_public, 0 )
		, entity_parameter_list(0)
		, parameter_list(param_list_)
	{}

	CsEntityFormalParameterList
		( cs_entity entity_type, CsFormalParameterList * param_list_ )
		: CsEntity ( entity_type, ac_public, 0 )
		, entity_parameter_list(0)
		, parameter_list(param_list_)
	{}

	virtual ~CsEntityFormalParameterList()
	{
		if ( entity_parameter_list )
		{
			entity_parameter_list->delete_nodes_and_values();
			delete entity_parameter_list;
		}
	}

	virtual void visit ( CsEntityVisitor * visitor )
	{
		visitor->visit ( this );
		if ( entity_parameter_list )
		{
			single_list1<CsEntityFormalParameter*>::node * next
				= & entity_parameter_list->first_;
			while ( next )
			{
				visitor->visit ( next->value_ );
				next = next->next_;
			}
		}
	}

	static inline void * operator new ( size_t size );
	static inline void operator delete ( void * address );

	CsFormalParameterList * parameter_list; // maybe 0 for property functions
	single_list1<CsEntityFormalParameter*> * entity_parameter_list;
};

DEFINE_ENTITY_ALLOCATOR(CsEntityFormalParameterList)

//------------------------------------------------------------------------------
// CsEntityMethodSpecifier
//------------------------------------------------------------------------------

struct CsEntityMethodSpecifier
	: public CsEntityFormalParametersSpecifier
{
	CsEntityMethodSpecifier()
	{}

	void delete_entities()
	{
		__super::delete_entities();
		return_type.delete_entities();
	}

	~CsEntityMethodSpecifier()
	{}

	CsEntityMethodSpecifier clone() const;

	bool equal ( const CsEntityMethodSpecifier & rhs ) const;


	static inline void * operator new ( size_t size );
	static inline void operator delete ( void * address );

	CsEntityTypeRef return_type;
	//cs_entity_method_calling_convention calling_convention;
};

DEFINE_ENTITY_ALLOCATOR(CsEntityMethodSpecifier)

//------------------------------------------------------------------------------
// CsEntityMethod
//------------------------------------------------------------------------------

struct CsEntityMethod : public CsEntity
{
	CsEntityMethod()
		: CsEntity ( ent_method, ac_public, 0 )
		, parameter_list(0)
		, block(0)
		, owner(0)
		, labels(0)
	{}

	CsEntityMethod ( WPHashString phs, cs_entity_access ac, CsNode * decl_ )
		: CsEntity ( ent_method, ac, phs )
		, decl(decl_)
		, parameter_list(0)
		, block(0)
		, owner(0)
		, labels(0)
	{}

	virtual ~CsEntityMethod()
	{
		delete parameter_list;
		delete block;
		specifier.delete_entities();
		delete labels;
	}

	virtual bool isSharedEntity() const
	{
		return true;
	}

	virtual void visit ( CsEntityVisitor * visitor )
	{
		visitor->visit ( this );
		if ( parameter_list )
			parameter_list->visit ( visitor );
		if ( block )
			block->visit ( visitor );
	}

	static inline void * operator new ( size_t size );
	static inline void operator delete ( void * address );

	CsEntityMethodSpecifier specifier;
	CsEntityFormalParameterList * parameter_list;
	CsEntityBlock * block; // this pointer used only for delete entity_block
	CsEntity * owner; // 0, property or event
	ScopeHashTable * labels; // contains pointers to CsNode
	CsNode * decl;
};

DEFINE_ENTITY_ALLOCATOR(CsEntityMethod)

//------------------------------------------------------------------------------
// CsEntityMethodImplementation
//------------------------------------------------------------------------------

struct CsEntityMethodImplementation
{
	CsEntityMethodImplementation()
		: interface_method(0)
		, implementation_method(0)
	{}

	CsEntityMethodImplementation
		( CsEntityMethod * interface_method_,
			CsEntityMethod * implementation_method_ )
		: interface_method(0)
		, implementation_method(implementation_method_)
	{}

	~CsEntityMethodImplementation()
	{}

	static inline void * operator new ( size_t size );
	static inline void operator delete ( void * address );

	CsEntityMethod * interface_method;
	CsEntityMethod * implementation_method;
};

DEFINE_ENTITY_ALLOCATOR(CsEntityMethodImplementation)

//------------------------------------------------------------------------------
// CsEntityType
//------------------------------------------------------------------------------

struct CsEntityClass;

struct CsEntityType : public CsScopeEntity
{
	CsEntityType ( cs_entity entity_type, cs_entity_access access, WPHashString phs, CsNode * node )
		: CsScopeEntity ( entity_type, access, phs, node )
		, base(0)
		, method_imlementations(0)
		, indexers(0)
		, derived_types(0)
	{}

	CsEntityType ( cs_entity entity_type, cs_entity_access access, WPHashString phs, CsNode * node, unsigned power_of_2 )
		: CsScopeEntity ( entity_type, access, phs, node, power_of_2 )
		, base(0)
		, method_imlementations(0)
		, indexers(0)
		, derived_types(0)
	{}

	~CsEntityType()
	{
		if ( method_imlementations )
		{
			delete method_imlementations;
		}
		if ( indexers )
		{
			indexers->delete_nodes();
			delete indexers;
		}
		if ( derived_types )
		{
			derived_types->delete_nodes();
			delete derived_types;
		}
	}

	void addMethodImplementation
		( CsEntityMethod * interface_method,
			CsEntityMethod * implementation_method )
	{
		if ( !method_imlementations )
		{
			method_imlementations
				= new value_single_list1<CsEntityMethodImplementation>
					( CsEntityMethodImplementation ( interface_method, implementation_method ) );
		}
		else
		{
			method_imlementations->push_back
				( CsEntityMethodImplementation ( interface_method, implementation_method ) );
		}
	}

	void addIndexer ( CsEntityProperty * indexer )
	{
		if ( !indexers )
		{
			indexers = new single_list1<CsEntityProperty*> ( indexer );
		}
		else
		{
			indexers->push_back ( indexer );
		}
	}

	static bool isBaseDerived ( CsEntityType * base, CsEntityType * derived );

	void addDerivedType ( CsEntityType * type )
	{
		if ( !derived_types )
		{
			derived_types = new single_list1<CsEntityType*> ( type );
		}
		else
		{
			derived_types->push_back ( type );
		}
	}

	virtual void deleteProjectEntities();
	void removeDerivedType ( CsEntityType * type );


	CsEntityClass * base;
	CsEntityTypeRef entity_typeref; // itself type representation in CsEntityTypeRef form
	value_single_list1<CsEntityMethodImplementation> * method_imlementations;
	single_list1<CsEntityProperty*> * indexers;
	single_list1<CsEntityType*> * derived_types;
};

//------------------------------------------------------------------------------
// CsEntityTypeWithInterfaces
//------------------------------------------------------------------------------

struct CsEntityTypeWithInterfaces : public CsEntityType
{
	CsEntityTypeWithInterfaces ( cs_entity entity_type, cs_entity_access access, WPHashString phs, CsNode * node )
		: CsEntityType ( entity_type, access, phs, node )
		, interfaces ( 0 )
	{}

	CsEntityTypeWithInterfaces ( cs_entity entity_type, cs_entity_access access, WPHashString phs, CsNode * node, unsigned power_of_2 )
		: CsEntityType ( entity_type, access, phs, node, power_of_2 )
		, interfaces ( 0 )
	{}

	virtual ~CsEntityTypeWithInterfaces()
	{
		if ( interfaces )
			interfaces->delete_nodes();
		delete interfaces;
	}

	virtual void deleteProjectEntities();


	single_list1<CsEntityInterface*> * interfaces;
};

//------------------------------------------------------------------------------
// CsEntityClass
//------------------------------------------------------------------------------

struct CsEntityClass : public CsEntityTypeWithInterfaces
{
	CsEntityClass ( WPHashString phs, cs_entity_access ac, CsNode * node )
		: CsEntityTypeWithInterfaces ( ent_class, ac, phs, node )
	{
		entity_typeref.type = et_class;
		entity_typeref.u.entity = this;
	}

	CsEntityClass ( WPHashString phs, cs_entity_access ac, CsNode * node, unsigned power_of_2 )
		: CsEntityTypeWithInterfaces ( ent_class, ac, phs, node, power_of_2 )
	{
		entity_typeref.type = et_class;
		entity_typeref.u.entity = this;
	}

	virtual ~CsEntityClass()
	{}

	virtual void visit ( CsEntityVisitor * visitor )
	{
		visitor->visit ( this );
		CsScopeEntity::visit ( visitor );
	}

	static inline void * operator new ( size_t size );
	static inline void operator delete ( void * address );
};

DEFINE_ENTITY_ALLOCATOR(CsEntityClass)

//------------------------------------------------------------------------------
// CsEntityStruct
//------------------------------------------------------------------------------

struct CsEntityStruct : public CsEntityTypeWithInterfaces
{
	CsEntityStruct ( WPHashString phs, cs_entity_access ac, CsNode * node )
		: CsEntityTypeWithInterfaces ( ent_struct, ac, phs, node )
	{
		entity_typeref.type = et_valuetype;
		entity_typeref.u.entity = this;
	}

	CsEntityStruct ( WPHashString phs, cs_entity_access ac, CsNode * node, unsigned power_of_2 )
		: CsEntityTypeWithInterfaces ( ent_struct, ac, phs, node, power_of_2 )
	{
		entity_typeref.type = et_valuetype;
		entity_typeref.u.entity = this;
	}

	virtual ~CsEntityStruct()
	{}

	virtual void visit ( CsEntityVisitor * visitor )
	{
		visitor->visit ( this );
		CsScopeEntity::visit ( visitor );
	}

	static inline void * operator new ( size_t size );
	static inline void operator delete ( void * address );
};

DEFINE_ENTITY_ALLOCATOR(CsEntityStruct)

//------------------------------------------------------------------------------
// CsEntityInterface
//------------------------------------------------------------------------------

struct CsEntityInterface : public CsEntityTypeWithInterfaces
{
	CsEntityInterface ( WPHashString phs, cs_entity_access ac, CsNode * node )
		: CsEntityTypeWithInterfaces ( ent_interface, ac, phs, node )
	{
		entity_typeref.type = et_class;
		entity_typeref.u.entity = this;
	}

	CsEntityInterface ( WPHashString phs, cs_entity_access ac, CsNode * node, unsigned power_of_2 )
		: CsEntityTypeWithInterfaces ( ent_interface, ac, phs, node, power_of_2 )
	{
		entity_typeref.type = et_class;
		entity_typeref.u.entity = this;
	}

	virtual ~CsEntityInterface()
	{}

	virtual void visit ( CsEntityVisitor * visitor )
	{
		visitor->visit ( this );
		CsScopeEntity::visit ( visitor );
	}

	static inline void * operator new ( size_t size );
	static inline void operator delete ( void * address );
};

DEFINE_ENTITY_ALLOCATOR(CsEntityInterface)

//------------------------------------------------------------------------------
// CsEntityEnum
//------------------------------------------------------------------------------

struct CsEntityEnum : public CsEntityType
{
	CsEntityEnum ( WPHashString phs, cs_entity_access ac, CsNode * node )
		: CsEntityType ( ent_enum, ac, phs, node )
	{
		entity_typeref.type = et_valuetype;
		entity_typeref.u.entity = this;
	}

	CsEntityEnum ( WPHashString phs, cs_entity_access ac, CsNode * node, unsigned power_of_2 )
		: CsEntityType ( ent_enum, ac, phs, node, power_of_2 )
	{
		entity_typeref.type = et_valuetype;
		entity_typeref.u.entity = this;
	}

	virtual ~CsEntityEnum()
	{
		base_type.delete_entities();
	}

	virtual void visit ( CsEntityVisitor * visitor )
	{
		visitor->visit ( this );
		CsScopeEntity::visit ( visitor );
	}

	static inline void * operator new ( size_t size );
	static inline void operator delete ( void * address );

	CsEntityTypeRef base_type;
};

DEFINE_ENTITY_ALLOCATOR(CsEntityEnum)

//------------------------------------------------------------------------------
// CsEntityConstant
//------------------------------------------------------------------------------

struct CsEntityConstant : public CsEntity
{
	CsEntityConstant ( WPHashString phs, cs_entity_access ac, CsNode * decl_ )
		: CsEntity ( ent_constant, ac, phs )
		, decl(decl_)
	{}

	virtual ~CsEntityConstant()
	{
		type.delete_entities();
	}

	virtual void visit ( CsEntityVisitor * visitor )
	{
		visitor->visit ( this );
	}

	static inline void * operator new ( size_t size );
	static inline void operator delete ( void * address );

	CsEntityTypeRef type;
	CsNode * decl;
};

DEFINE_ENTITY_ALLOCATOR(CsEntityConstant)

//------------------------------------------------------------------------------
// CsEntityVariable
//------------------------------------------------------------------------------

struct CsEntityVariable : public CsEntity
{
	CsEntityVariable ( WPHashString phs, cs_entity_access ac, CsVariableDeclarator * decl_ )
		: CsEntity ( ent_variable, ac, phs )
		, decl(decl_)
	{}

	virtual ~CsEntityVariable()
	{
		type.delete_entities();
	}

	virtual void visit ( CsEntityVisitor * visitor )
	{
		visitor->visit ( this );
	}

	static inline void * operator new ( size_t size );
	static inline void operator delete ( void * address );

	CsEntityTypeRef type;
	CsVariableDeclarator * decl;
};

DEFINE_ENTITY_ALLOCATOR(CsEntityVariable)

//------------------------------------------------------------------------------
// CsEntityBlockVariable
//------------------------------------------------------------------------------

struct CsEntityBlockVariable : public CsEntity
{
	CsEntityBlockVariable ( WPHashString phs, CsNode * decl_ )
		: CsEntity ( ent_block_variable, ac_public, phs )
		, decl(decl_)
		, block(0)
	{}

	virtual ~CsEntityBlockVariable()
	{
		if ( block )
		{
			delete block;
		}
	}

	virtual void visit ( CsEntityVisitor * visitor )
	{
		visitor->visit ( this );
		if ( block )
			block->visit ( visitor );
	}

	static inline void * operator new ( size_t size );
	static inline void operator delete ( void * address );

	CsNode * decl;
	CsEntity * block;
};

DEFINE_ENTITY_ALLOCATOR(CsEntityBlockVariable)

//------------------------------------------------------------------------------
// CsEntityDelegate
//------------------------------------------------------------------------------

struct CsEntityDelegate : public CsEntityType
{
	CsEntityDelegate ( WPHashString phs, cs_entity_access ac, CsNode * node )
		: CsEntityType ( ent_delegate, ac, phs, node )
		, parameter_list(0)
	{
		entity_typeref.type = et_class;
		entity_typeref.u.entity = this;
	}

	CsEntityDelegate ( WPHashString phs, cs_entity_access ac, CsNode * node, unsigned power_of_2 )
		: CsEntityType ( ent_delegate, ac, phs, node, power_of_2 )
		, parameter_list(0)
	{
		entity_typeref.type = et_class;
		entity_typeref.u.entity = this;
	}

	virtual ~CsEntityDelegate()
	{
		delete parameter_list;
		specifier.delete_entities();
	}

	virtual void visit ( CsEntityVisitor * visitor )
	{
		visitor->visit ( this );
		if ( parameter_list )
			parameter_list->visit ( visitor );
	}

	static inline void * operator new ( size_t size );
	static inline void operator delete ( void * address );

	CsEntityFormalParameterList * parameter_list; // this pointer used only for delete entity_block
	CsEntityMethodSpecifier specifier;
};

DEFINE_ENTITY_ALLOCATOR(CsEntityDelegate)

//------------------------------------------------------------------------------
// CsEntityProperty
//------------------------------------------------------------------------------

struct CsEntityProperty : public CsEntity
{
	CsEntityProperty ( WPHashString phs, cs_entity_access ac, CsNode * decl_ )
		: CsEntity ( ent_property, ac, phs )
		, decl(decl_)
		, parameter_list(0)
		, getter(0)
		, setter(0)
		, others(0)
	{}

	virtual ~CsEntityProperty()
	{
		entity_type.delete_entities();
		specifier.delete_entities();
		delete parameter_list;
		if ( others )
			others->delete_nodes();
		delete others;
	}

	virtual bool isSharedEntity() const
	{
		return true;
	}

	virtual void visit ( CsEntityVisitor * visitor )
	{
		visitor->visit ( this );
		if ( parameter_list )
			parameter_list->visit ( visitor );
	}

	static inline void * operator new ( size_t size );
	static inline void operator delete ( void * address );

	CsNode * decl;
	CsEntityTypeRef entity_type;
	CsEntityFormalParametersSpecifier specifier;
	CsEntityFormalParameterList * parameter_list;
	CsEntityMethod * getter;
	CsEntityMethod * setter;
	single_list1<CsEntityMethod*> * others;
};

DEFINE_ENTITY_ALLOCATOR(CsEntityProperty)

//------------------------------------------------------------------------------
// CsEntityEvent
//------------------------------------------------------------------------------

struct CsEntityEvent : public CsEntity
{
	CsEntityEvent ( WPHashString phs, cs_entity_access ac, CsEventDeclarator * declarator_ )
		: CsEntity ( ent_event, ac, phs )
		, declarator(declarator_)
		, add(0)
		, remove(0)
		, fire(0)
		, others(0)
	{}

	virtual ~CsEntityEvent()
	{
		entity_type.delete_entities();
		if ( others )
			others->delete_nodes();
		delete others;
	}

	virtual void visit ( CsEntityVisitor * visitor )
	{
		visitor->visit ( this );
	}

	static inline void * operator new ( size_t size );
	static inline void operator delete ( void * address );

	CsEventDeclarator * declarator;
	CsEntityTypeRef entity_type;
	CsEntityMethod * add;
	CsEntityMethod * remove;
	CsEntityMethod * fire;
	single_list1<CsEntityMethod*> * others;
};

DEFINE_ENTITY_ALLOCATOR(CsEntityEvent)

//------------------------------------------------------------------------------
// CsEntityLocalConstant
//------------------------------------------------------------------------------

struct CsEntityLocalConstant : public CsEntity
{
	CsEntityLocalConstant ( WPHashString phs, CsLocalConstantDeclarator * decl_ )
		: CsEntity ( ent_local_constant, ac_public, phs )
		, decl(decl_)
	{}

	virtual ~CsEntityLocalConstant()
	{}

	virtual void visit ( CsEntityVisitor * visitor )
	{
		visitor->visit ( this );
	}

	static inline void * operator new ( size_t size );
	static inline void operator delete ( void * address );

  CsLocalConstantDeclarator * decl;
};

DEFINE_ENTITY_ALLOCATOR(CsEntityLocalConstant)

//------------------------------------------------------------------------------
// CsEntityLocalVariable
//------------------------------------------------------------------------------

struct CsEntityLocalVariable : public CsEntity
{
	CsEntityLocalVariable ( WPHashString phs, CsNode * decl_ )
		: CsEntity ( ent_local_variable, ac_public, phs )
		, decl(decl_)
	{}

	virtual ~CsEntityLocalVariable()
	{}

	virtual void visit ( CsEntityVisitor * visitor )
	{
		visitor->visit ( this );
	}

	static inline void * operator new ( size_t size );
	static inline void operator delete ( void * address );

	CsNode * decl;
};

DEFINE_ENTITY_ALLOCATOR(CsEntityLocalVariable)

//------------------------------------------------------------------------------


#endif // _CS_ENTITY_H_
